/*
 * blanco Framework
 * Copyright (C) 2004-2006 IGA Tosiki
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 */
package blanco.cg.transformer.java;

import java.util.ArrayList;
import java.util.List;

import blanco.cg.BlancoCgSupportedLang;
import blanco.cg.util.BlancoCgLineUtil;
import blanco.cg.valueobject.BlancoCgException;
import blanco.cg.valueobject.BlancoCgLangDoc;
import blanco.cg.valueobject.BlancoCgMethod;
import blanco.cg.valueobject.BlancoCgParameter;
import blanco.cg.valueobject.BlancoCgSourceFile;
import blanco.commons.util.BlancoNameUtil;
import blanco.commons.util.BlancoStringUtil;

/**
 * BlancoCgMethod\[XR[hɓWJ܂B
 * 
 * ̃NXblancoCg̃o[IuWFNg\[XR[hgXtH[}[̌ʂ̓WJ@\łB
 * 
 * @author IGA Tosiki
 */
class BlancoCgMethodJavaSourceExpander {
    /**
     * ̃NXΏۂƂvO~OB
     */
    protected static final int TARGET_LANG = BlancoCgSupportedLang.JAVA;

    /**
     * Ń\bhWJ܂B
     * 
     * @param cgMethod
     *            ΏۂƂȂ郁\bhB
     * @param argSourceFile
     *            \[Xt@CB
     * @param argSourceLines
     *            o͐sXgB
     * @param argIsInterface
     *            C^tF[XǂBNX̏ꍇɂfalseBC^tF[X̏ꍇɂtrueB
     */
    public void transformMethod(final BlancoCgMethod cgMethod,
            final BlancoCgSourceFile argSourceFile,
            final List<java.lang.String> argSourceLines,
            final boolean argIsInterface) {
        if (BlancoStringUtil.null2Blank(cgMethod.getName()).length() == 0) {
            throw new IllegalArgumentException("\bh̖OɓK؂Ȓlݒ肳Ă܂B");
        }
        if (cgMethod.getReturn() == null) {
            // ͂肦܂Bvoid̏ꍇɂnullw肳̂łB
        }

        // st^B
        argSourceLines.add("");

        prepareExpand(cgMethod, argSourceFile);

        // 񂪈ꎮ̂ŁA\[XR[h̎ۂ̓WJs܂B

        //  LangDoc\[XR[h`ɓWJB
        new BlancoCgLangDocJavaSourceExpander().transformLangDoc(cgMethod
                .getLangDoc(), argSourceLines);

        // Ame[VWJB
        expandAnnotationList(cgMethod, argSourceLines);

        // \bh̖{̕WJB
        expandMethodBody(cgMethod, argSourceLines, argIsInterface);
    }

    /**
     * \[XR[hWJɐ旧AKvȏ̎Ws܂B
     * 
     * @param cgMethod
     *            \bhIuWFNgB
     * @param argSourceFile
     *            \[Xt@CB
     */
    private void prepareExpand(final BlancoCgMethod cgMethod,
            final BlancoCgSourceFile argSourceFile) {
        // ŏɃ\bhLangDocɓWJB
        if (cgMethod.getLangDoc() == null) {
            // LangDocw̏ꍇɂ͂瑤ŃCX^X𐶐B
            cgMethod.setLangDoc(new BlancoCgLangDoc());
        }
        if (cgMethod.getLangDoc().getParameterList() == null) {
            cgMethod.getLangDoc().setParameterList(
                    new ArrayList<blanco.cg.valueobject.BlancoCgParameter>());
        }
        if (cgMethod.getLangDoc().getThrowList() == null) {
            cgMethod.getLangDoc().setThrowList(
                    new ArrayList<blanco.cg.valueobject.BlancoCgException>());
        }
        if (cgMethod.getLangDoc().getTitle() == null) {
            cgMethod.getLangDoc().setTitle(cgMethod.getDescription());
        }

        for (BlancoCgParameter cgParameter : cgMethod.getParameterList()) {
            // importɌ^ǉB
            argSourceFile.getImportList().add(cgParameter.getType().getName());

            // hLgɃp[^ǉB
            cgMethod.getLangDoc().getParameterList().add(cgParameter);
        }

        if (cgMethod.getReturn() != null) {
            // importɌ^ǉB
            argSourceFile.getImportList().add(
                    cgMethod.getReturn().getType().getName());

            // hLgreturnǉB
            cgMethod.getLangDoc().setReturn(cgMethod.getReturn());
        }

        // OɂLangDoc\̂ɓWJ
        for (BlancoCgException cgException : cgMethod.getThrowList()) {
            // importɌ^ǉB
            argSourceFile.getImportList().add(cgException.getType().getName());

            // hLgɗOǉB
            cgMethod.getLangDoc().getThrowList().add(cgException);
        }
    }

    /**
     * \bh̖{̕WJ܂B
     * 
     * @param cgMethod
     *            \bhIuWFNgB
     * @param argSourceLines
     *            \[XR[hB
     * @param argIsInterface
     *            C^tF[XƂēWJ邩ǂB
     */
    private void expandMethodBody(final BlancoCgMethod cgMethod,
            final List<java.lang.String> argSourceLines,
            final boolean argIsInterface) {
        final StringBuffer buf = new StringBuffer();

        if (BlancoStringUtil.null2Blank(cgMethod.getAccess()).length() > 0) {
            if (argIsInterface && cgMethod.getAccess().equals("public")) {
                // C^tF[Xpublic̏ꍇɂ͏o͂}܂B
                // Checkstyle΍ƂȂ܂B
            } else {
                if (cgMethod.getStaticInitializer() == false) {
                    buf.append(cgMethod.getAccess() + " ");
                }
            }
        }

        if (cgMethod.getAbstract() && argIsInterface == false) {
            // C^tF[X̏ꍇɂ abstract͕t^܂B
            buf.append("abstract ");
        }
        if (cgMethod.getStatic()) {
            buf.append("static ");
        }
        if (cgMethod.getStaticInitializer()) {
            buf.append("static");
        }
        if (cgMethod.getFinal() && argIsInterface == false) {
            // C^tF[X̏ꍇɂ final͕t^܂B
            buf.append("final ");
        }

        if (cgMethod.getConstructor()) {
            // RXgN^̏ꍇɂ́A߂l݂͑܂B
            // ̂߁Ał͉o͂܂B
        } else if (cgMethod.getStaticInitializer()) {
            // static initializer ̏ꍇɂ́A߂l݂͑܂B
            // ̂߁Ał͉o͂܂B
        } else {
            if (cgMethod.getReturn() != null
                    && cgMethod.getReturn().getType() != null) {
                buf.append(BlancoCgTypeJavaSourceExpander.toTypeString(cgMethod
                        .getReturn().getType())
                        + " ");
            } else {
                buf.append("void ");
            }
        }

        if (cgMethod.getStaticInitializer() == false) {
            buf.append(cgMethod.getName() + "(");
            for (int index = 0; index < cgMethod.getParameterList().size(); index++) {
                final BlancoCgParameter cgParameter = cgMethod
                        .getParameterList().get(index);
                if (cgParameter.getType() == null) {
                    throw new IllegalArgumentException("\bh["
                            + cgMethod.getName() + "]̃p[^["
                            + cgParameter.getName() + "]Ɍ^null^܂B");
                }

                if (index != 0) {
                    buf.append(", ");
                }

                if (cgParameter.getFinal()) {
                    buf.append("final ");
                }
                buf.append(BlancoCgTypeJavaSourceExpander
                        .toTypeString(cgParameter.getType()));
                buf.append(" ");
                buf.append(cgParameter.getName());
            }
            buf.append(")");
        }

        // OX[WJB
        expandThrowList(cgMethod, buf);

        if (cgMethod.getAbstract() || argIsInterface) {
            // ۃ\bh܂̓C^tF[X̏ꍇɂ́A\bh̖{̂WJ܂B
            buf.append(BlancoCgLineUtil.getTerminator(TARGET_LANG));
            argSourceLines.add(buf.toString());
        } else {
            // \bhubN̊JnB
            buf.append(" {");

            // łAsmB
            argSourceLines.add(buf.toString());

            // eNX\bhs@\̓WJB
            if (BlancoStringUtil.null2Blank(cgMethod.getSuperclassInvocation())
                    .length() > 0) {
                // super() Ȃǂ܂܂܂B
                argSourceLines.add(cgMethod.getSuperclassInvocation()
                        + BlancoCgLineUtil.getTerminator(TARGET_LANG));
            }

            // p[^̔null̓WJB
            expandParameterCheck(cgMethod, argSourceLines);

            // sWJ܂B
            expandLineList(cgMethod, argSourceLines);

            // \bhubN̏IB
            argSourceLines.add("}");
        }
    }

    /**
     * OX[WJ܂B
     * 
     * @param cgMethod
     *            \bhB
     * @param buf
     *            o̓obt@B
     */
    private void expandThrowList(final BlancoCgMethod cgMethod,
            final StringBuffer buf) {
        for (int index = 0; index < cgMethod.getThrowList().size(); index++) {
            final BlancoCgException cgException = cgMethod.getThrowList().get(
                    index);
            if (index == 0) {
                buf.append(" throws ");
            } else {
                buf.append(", ");
            }
            // hLgɂẮAblancoCgTypeɊւ鋤ʏ𗘗p邱Ƃ͂ł܂B
            // ʂɋLqs܂B
            buf.append(BlancoNameUtil.trimJavaPackage(cgException.getType()
                    .getName()));
        }
    }

    /**
     * Ame[VWJ܂B
     * 
     * @param cgMethod
     *            \bhB
     * @param argSourceLines
     *            \[XR[hB
     */
    private void expandAnnotationList(final BlancoCgMethod cgMethod,
            final List<java.lang.String> argSourceLines) {
        if (cgMethod.getOverride()) {
            // Javał override̓Ame[Vŕ\܂B
            argSourceLines.add("@Override");
        }

        for (String strAnnotation : cgMethod.getAnnotationList()) {
            // JavaAnnotation @ Lq܂B
            argSourceLines.add("@" + strAnnotation);
        }
    }

    /**
     * p[^̔null̓WJB
     * 
     * @param cgMethod
     *            \bhB
     * @param argSourceLines
     *            \[XR[hB
     */
    private void expandParameterCheck(final BlancoCgMethod cgMethod,
            final List<java.lang.String> argSourceLines) {
        boolean isProcessed = false;
        for (BlancoCgParameter cgParameter : cgMethod.getParameterList()) {
            if (cgParameter.getNotnull()) {
                isProcessed = true;

                argSourceLines.add(BlancoCgLineUtil.getIfBegin(TARGET_LANG,
                        cgParameter.getName() + " == null"));
                argSourceLines.add("throw new IllegalArgumentException(\"\bh["
                        + cgMethod.getName() + "]̃p[^["
                        + cgParameter.getName()
                        + "]null^܂BÃp[^null^邱Ƃ͂ł܂B\");");
                argSourceLines.add(BlancoCgLineUtil.getIfEnd(TARGET_LANG));
            }
        }

        if (isProcessed) {
            // p[^`FbNWJꂽꍇɂ͋s}܂B
            argSourceLines.add("");
        }
    }

    /**
     * sWJ܂B
     * 
     * @param cgMethod
     *            \bhB
     * @param argSourceLines
     *            o͍sXgB
     */
    private void expandLineList(final BlancoCgMethod cgMethod,
            final List<java.lang.String> argSourceLines) {
        for (String strLine : cgMethod.getLineList()) {
            argSourceLines.add(strLine);
        }
    }
}
